html,body {
  height:100%;
}

.main {
  position: relative;
}

#top-bar {
  background: #039;
}

/* Top bar nav items */
.navbar .nav-item .nav-link {
  color: var(--bg-primary) !important;
}

#logo-col {
  justify-content: center;
  display: flex;
}

#project-name {
  font-family: monospace !important;
  font-weight: 900 !important;
  font-size: 24px;
}

#experiments-title {
  margin-top:20px;
}

/* Dash DataTable - Dark theme */

.dash-cell {
  padding-left: 12px !important;
  padding-right: 12px !important;
}

.column-header--sort {
  padding-left: 12px !important;
  padding-right: 8px !important;
}

.dash-filter--case {
  margin-left: 12px !important;
}

.dash-filter input[type=text] {
  color: var(--bs-body-color) !important;
  text-align: left !important;
  margin-left: 12px !important;
}

.first-page,
.previous-page,
.next-page,
.last-page
{
  color: var(--bs-body-color) !important;
  background-color:var(--bs-secondary);
}
.current-page {
  color: var(--bs-body-color) !important;
  background-color: var(--bs-secondary);
}

#no-paging-spacer {
  height:40px
}

input.current-page::placeholder {
  color: var(--bs-body-color) !important;
}

/* Tab colors */
.nav .nav-item .active.nav-link {
  background-color: rgb(var(--bs-dark-rgb)) !important;
}

.nav .nav-item .nav-link {
  background-color: #2a2a2a !important;
}


.tab-content {
  width:100%;
  min-height:400px !important;
  background-color: rgb(var(--bs-dark-rgb)); /*#333333;*/
}

.tab-placeholder-container {
  width:100%;
  min-height:400px !important;
  display: table;
  overflow: hidden;
}

.tab-placeholder-text {
  display: table-cell;
  vertical-align: middle;
  text-align: center;
  color: #999999;
  width: 500px;
  height: 450px;
}

.add-button-col {
  display: table;
  overflow: hidden;
}

.add-button-col-container {
  display: table-cell;
  vertical-align: middle;
}

#add-button {
  margin-top: 45px;
  font-size:12px;
}

#remove-title {
  margin-top:40px;
  font-size:12px;
}

.remove-button {
  width: 70px;
  margin-top: 4px;
  margin-bottom: 4px;
  border-color: #666666;
  font-size:12px;
}

.remove-button:hover {
  border-color: #999999;
}

.modal-backdrop.show {
  opacity: 0.8;
}

#failed-plotting-alert {
  position: absolute;
  top: 20px;
  right: 20px;
  z-index: 100;
  opacity: 95%;
}

#aggregate-container {
  position: relative;
}

#aggregate-tabs .nav-item {
  width:154px;
}

#copy-data-button {
  position: absolute;
  top: 3px;
  left: 240px;
  font-size:12px;
}
